"""
Core engine abstractions
"""

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel


class SelectEngine(ABC):
    """Abstract base class for database query engines"""

    @abstractmethod
    def select(self, query: str, *params: Any) -> list[Any]:
        """Execute a SELECT query and return results"""
        pass

    @abstractmethod
    def select_count(self, query: str, *params: Any) -> int:
        """Execute a COUNT query and return the count"""
        pass


class QueryContext(BaseModel):
    """
    Context object passed to middlewares, containing all query-related data
    that can be modified by middleware.
    """

    select_query: str
    select_params: list[Any] | None = None
    count_query: str | None = None
    count_params: list[Any] | None = None
    filters: BaseModel
    limit: int
    offset: int
    dependency_results: dict[str, Any]



from typing import Any, Callable

from pydantic import BaseModel


def get_parsed_count_params(
    filters: BaseModel,
    count_param: Callable[[Any], list[Any]] | list[str] | None,
    base_param: Callable[[Any], list[Any]] | list[str] | None,
) -> list[Any]:
    param = count_param if count_param is not None else base_param
    if param is None:
        return []
    if callable(param):
        return param(filters)
    if isinstance(param, list):
        return [_get_attr(filters, attr_name) for attr_name in param]
    return []


def get_parsed_select_params(
    filters: BaseModel,
    limit: int,
    offset: int,
    select_param: Callable[[Any, int, int], list[Any]] | list[str] | None,
    base_param: Callable[[Any], list[Any]] | list[str] | None,
) -> list[Any]:
    if select_param is None and base_param is None:
        return [limit, offset]
    if select_param is not None:
        if callable(select_param):
            return select_param(filters, limit, offset)
        if isinstance(select_param, list):
            parsed_param = []
            for attr_name in select_param:
                if attr_name == "limit":
                    parsed_param.append(limit)
                    continue
                if attr_name == "offset":
                    parsed_param.append(offset)
                    continue
                parsed_param.append(_get_attr(filters, attr_name))
            return parsed_param
        return [limit, offset]
    if base_param is not None:
        if callable(base_param):
            return base_param(filters) + [limit, offset]
        if isinstance(base_param, list):
            return [_get_attr(filters, attr_name) for attr_name in base_param] + [
                limit,
                offset,
            ]
    return [limit, offset]


def _get_attr(obj: Any, attr_names: str) -> Any:
    attr = obj
    for attr_name in attr_names.split("."):
        if not hasattr(attr, attr_name):
            raise ValueError(f"{obj} has no attribute {attr_name}")
        attr = getattr(attr, attr_name)
    return attr



"""
Core abstractions for DBAnu
"""

from dbanu.core.engine import QueryContext, SelectEngine
from dbanu.core.middleware import Middleware
from dbanu.core.response import create_select_response_model

__all__ = ["SelectEngine", "QueryContext", "Middleware", "create_select_response_model"]



import inspect
import traceback
from typing import Any, Callable, Type, TypeVar

from fastapi import Body, Depends, FastAPI, HTTPException, Request
from pydantic import BaseModel, create_model

from dbanu.api.dependencies import create_wrapped_fastapi_dependencies
from dbanu.core.engine import QueryContext, SelectEngine
from dbanu.core.middleware import (
    Middleware,
    create_middleware_chain,
    get_combined_middlewares,
    validate_middlewares,
)
from dbanu.core.response import create_select_response_model
from dbanu.utils.pagination import calculate_union_pagination
from dbanu.utils.param import get_parsed_count_params, get_parsed_select_params
from dbanu.utils.string import to_var_name

Filter = TypeVar("Filter", bound=BaseModel)


class SelectSource(BaseModel):
    """Configuration for a single data source in union queries"""

    model_config = {"arbitrary_types_allowed": True}
    query_engine: SelectEngine
    select_query: str | Callable[[BaseModel], str]
    select_param: Callable[[BaseModel, int, int], list[Any]] | list[str] | None = None
    count_query: str | Callable[[BaseModel], str]
    count_param: Callable[[BaseModel], list[Any]] | list[str] | None = None
    param: Callable[[BaseModel], list[Any]] | list[str] | None = None
    middlewares: list[Middleware] | None = None

    def __init__(self, **data):
        super().__init__(**data)
        # Validate that all middlewares are async functions
        validate_middlewares(self.middlewares)


def serve_union(
    app: FastAPI,
    sources: dict[str, SelectSource],
    path: str = "/get",
    methods: list[str] | None = None,
    filter_model: Type[BaseModel] | None = None,
    data_model: Type[BaseModel] | None = None,
    response_model: Type[BaseModel] | None = None,
    dependencies: list[Any] | None = None,
    middlewares: list[Middleware] | None = None,
    source_priority: list[str] | None = None,
    name: str | None = None,
    summary: str | None = None,
    description: str | None = None,
    default_limit: int | None = None,
):
    """
    Create a union endpoint that combines results from multiple sources
    with priority-based pagination
    """
    methods = ["GET"] if methods is None else [m.upper() for m in methods]
    var_name = to_var_name(name, path)
    if filter_model is None:
        filter_model = create_model(
            "FilterModel" if var_name is None else f"{var_name.capitalize()}Filter",
            limit=(int, default_limit if default_limit is not None else 100),
            offset=(int, 0),
            priority=(str | None, None),
        )
    if response_model is None:
        response_model = create_select_response_model(
            "ResponseModel" if var_name is None else f"{var_name.capitalize()}Response",
            data_model,
        )
    wrapped_dependencies = create_wrapped_fastapi_dependencies(dependencies)
    # Validate that all middlewares are async functions
    validate_middlewares(middlewares)

    # Create the actual handler function
    async def handle_request(
        request: Request,
        filter_data: filter_model,  # type: ignore
    ):
        """Union handler"""
        # Get params from filters
        limit = getattr(filter_data, "limit", 100)
        offset = getattr(filter_data, "offset", 0)
        priority = getattr(filter_data, "priority", None)
        # Extract dependency results from request state
        dependency_results = {}
        if request and hasattr(request.state, "dependency_results"):
            dependency_results = request.state.dependency_results
        priority_list = _get_priority_list(priority, source_priority, sources)
        # Step 1: Get total count from each source
        source_counts = {}
        total_count = 0
        for source_name in priority_list:
            source = sources[source_name]
            count_query_str = (
                source.count_query(filter_data)
                if callable(source.count_query)
                else source.count_query
            )
            parsed_count_params = get_parsed_count_params(
                filter_data, source.count_param, source.param
            )
            count_context = QueryContext(
                select_query="",
                select_params=[],
                count_query=count_query_str,
                count_params=parsed_count_params,
                filters=filter_data,
                limit=0,
                offset=0,
                dependency_results=dependency_results,
            )
            count_processor = _create_count_processor(source.query_engine)
            handler = create_middleware_chain(source.middlewares, count_processor)
            source_count = await handler(count_context)
            source_counts[source_name] = source_count
            total_count += source_count
        # Step 2: Calculate which records to fetch from each source
        fetch_plan = calculate_union_pagination(
            source_counts, priority_list, limit, offset
        )
        # Step 3: Fetch data according to the plan
        final_data = []
        for source_name, (source_limit, source_offset) in fetch_plan.items():
            source = sources[source_name]
            # Build select parameters for this specific source
            select_query_str = (
                source.select_query(filter_data)
                if callable(source.select_query)
                else source.select_query
            )
            parsed_select_params = get_parsed_select_params(
                filter_data,
                source_limit,
                source_offset,
                source.select_param,
                source.param,
            )
            # Create QueryContext for this source
            select_context = QueryContext(
                select_query=select_query_str,
                select_params=parsed_select_params,
                count_query="",
                count_params=[],
                filters=filter_data,
                limit=source_limit,
                offset=source_offset,
                dependency_results=dependency_results,
            )
            select_processor = _create_select_processor(source.query_engine)
            handler = create_middleware_chain(
                get_combined_middlewares(middlewares, source.middlewares),
                select_processor,
            )
            source_data = await handler(select_context)
            # Add the data to final result
            final_data.extend(source_data)
        return response_model(data=final_data, count=total_count)

    # Register routes based on methods
    if "GET" in methods:

        @app.get(
            path,
            response_model=response_model,
            dependencies=wrapped_dependencies,
            name=None if name is None else f"{name}Get",
            summary=summary,
            description=description,
        )
        async def get_handler(
            request: Request,
            filters: filter_model = Depends(),  # type: ignore
        ):
            """Union route (generated by dbAnu)"""
            try:
                return await handle_request(request, filters)
            except Exception as e:
                traceback.print_exc()
                raise HTTPException(500, f"{e}")

    # Register other methods (POST, PUT, etc.)
    other_methods = [m for m in methods if m != "GET"]
    if other_methods:
        # Create a single route for all non-GET methods
        @app.api_route(
            path,
            methods=other_methods,
            response_model=response_model,
            dependencies=wrapped_dependencies,
            name=name,
            summary=summary,
            description=description,
        )
        async def non_get_handler(
            request: Request,
            filters: filter_model = Body(),  # type: ignore
        ):
            """Union route (generated by dbAnu)"""
            try:
                return await handle_request(request, filters)
            except Exception as e:
                traceback.print_exc()
                raise HTTPException(500, f"{e}")


def _get_priority_list(
    source_priority_str: str | None,
    default_source_priority: list[str] | None,
    sources: dict[str, SelectSource],
) -> list:
    if source_priority_str is not None:
        return [
            source.strip()
            for source in source_priority_str.split(",")
            if source.strip() != ""
            if source in sources
        ]
    if default_source_priority is not None:
        return default_source_priority
    return list(sources.keys())


def _create_select_processor(query_engine: SelectEngine):
    """Create a select processor for the middleware chain"""

    async def process_select(context: QueryContext) -> list[Any]:
        select_params = context.select_params or []
        # Check if select method is a coroutine
        select_method = query_engine.select
        if inspect.iscoroutinefunction(select_method):
            return await select_method(context.select_query, *select_params)
        return select_method(context.select_query, *select_params)

    return process_select


def _create_count_processor(query_engine: SelectEngine):
    """Create a count processor for the middleware chain"""

    async def process_count(context: QueryContext) -> int:
        count_params = context.count_params or []
        if context.count_query is None:
            raise ValueError(f"select_count is not defined at this context: {context}")
        select_count_method = query_engine.select_count
        if inspect.iscoroutinefunction(select_count_method):
            return await select_count_method(context.count_query, *count_params)
        return select_count_method(context.count_query, *count_params)

    return process_count



"""
Union pagination utilities
"""

from typing import Dict, Tuple


def calculate_union_pagination(
    source_counts: Dict[str, int], priority: list[str], limit: int, offset: int
) -> Dict[str, Tuple[int, int]]:
    """
    Calculate which records to fetch from each source for union pagination

    Args:
        source_counts: Dictionary mapping source names to their record counts
        priority: List of source names in priority order
        limit: Total number of records to fetch
        offset: Starting offset in the combined dataset

    Returns:
        Dictionary mapping source names to (limit, offset) tuples
    """
    fetch_plan = {}
    remaining_limit = limit
    current_offset = offset

    for source_name in priority:
        source_count = source_counts[source_name]

        # Skip this source if offset is beyond its records
        if current_offset >= source_count:
            current_offset -= source_count
            continue

        # Calculate how many records to fetch from this source
        source_limit = min(remaining_limit, source_count - current_offset)

        if source_limit > 0:
            fetch_plan[source_name] = (source_limit, current_offset)
            remaining_limit -= source_limit
            # Reset offset for next source
            current_offset = 0

        # Stop if we've reached the limit
        if remaining_limit <= 0:
            break

    return fetch_plan



"""
PostgreSQL query engine
"""

from typing import Any

import psycopg2

from dbanu.core.engine import SelectEngine


class PostgreSQLQueryEngine(SelectEngine):
    """
    A PostgreSQL query engine that connects to a PostgreSQL database.
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 5432,
        database: str = "postgres",
        user: str = "postgres",
        password: str = "",
    ):
        self._host = host
        self._port = port
        self._database = database
        self._user = user
        self._password = password
        self._connection_params = {
            "host": host,
            "port": port,
            "database": database,
            "user": user,
            "password": password,
        }

    def _get_connection(self):
        """Get a connection to the PostgreSQL database."""
        return psycopg2.connect(**self._connection_params)

    def select(self, query: str, *params: Any) -> list[Any]:
        """
        Execute a SELECT query against the PostgreSQL database.
        """
        conn = self._get_connection()
        cursor = conn.cursor()
        try:
            # Execute the query with parameters
            if params:
                cursor.execute(query, params)
            else:
                cursor.execute(query)
            # Fetch results
            results = cursor.fetchall()
            # Convert to list of dictionaries for Pydantic compatibility
            if results and len(results) > 0:
                # Get column names
                column_names = [description[0] for description in cursor.description]
                # Convert each row to a dictionary
                dict_results = []
                for row in results:
                    dict_results.append(dict(zip(column_names, row)))
                return dict_results
            else:
                return []
        except Exception as e:
            raise e
        finally:
            cursor.close()
            conn.close()

    def select_count(self, query: str, *params: Any) -> int:
        """
        Execute a COUNT query against the PostgreSQL database.
        """
        conn = self._get_connection()
        cursor = conn.cursor()
        try:
            # Execute the query with parameters
            if params:
                cursor.execute(query, params)
            else:
                cursor.execute(query)
            # Fetch the count result
            result = cursor.fetchone()
            return result[0] if result else 0
        except Exception as e:
            raise e
        finally:
            cursor.close()
            conn.close()



"""
Response models for DBAnu
"""

from typing import Any

from pydantic import BaseModel, create_model


def create_select_response_model(
    model_name: str, data_model: type[BaseModel] | None = None
):
    """Create a response model for select endpoints"""
    actual_data_model = (
        create_model(f"{model_name}Data") if data_model is None else data_model
    )
    return create_model(
        model_name, data=(list[actual_data_model] | list[Any], ...), count=(int, ...)
    )



"""
Dependency injection utilities for DBAnu
"""

import inspect
from typing import Any

from fastapi import Depends, Request


def create_wrapped_fastapi_dependencies(dependencies: list[Any] | None):
    """Create wrapped dependencies that store results in request state"""
    if dependencies is None:
        return []
    wrapped_dependencies = []
    for dep in dependencies:
        # Create a closure-safe wrapper for each dependency
        def create_wrapped_dependency(original_dep):
            async def wrapped_dependency(request: Request):
                # Extract the actual dependency function from Depends object
                if hasattr(original_dep, "dependency"):
                    dependency_func = original_dep.dependency
                else:
                    dependency_func = original_dep

                # Execute the dependency
                if inspect.iscoroutinefunction(dependency_func):
                    result = await dependency_func()
                else:
                    result = dependency_func()

                # Store the result in request state
                if not hasattr(request.state, "dependency_results"):
                    request.state.dependency_results = {}

                # Use the dependency function name as key
                dep_name = dependency_func.__name__
                request.state.dependency_results[dep_name] = result
                return result

            return wrapped_dependency

        wrapped_dependencies.append(Depends(create_wrapped_dependency(dep)))

    return wrapped_dependencies



def to_var_name(text: str | None, *alternatives: str | None) -> str | None:
    for alternative in [text, *alternatives]:
        if alternative is None:
            continue
        var_name = "".join(
            c for c in alternative.title().replace(" ", "") if c.isalnum()
        ).lstrip("0123456789")
        if var_name != "":
            return var_name
    return None



import inspect
import traceback
from typing import Any, Callable, Type, TypeVar

from fastapi import Body, Depends, FastAPI, HTTPException, Request
from pydantic import BaseModel, create_model

from dbanu.api.dependencies import create_wrapped_fastapi_dependencies
from dbanu.core.engine import QueryContext, SelectEngine
from dbanu.core.middleware import (
    Middleware,
    create_middleware_chain,
    validate_middlewares,
)
from dbanu.core.response import create_select_response_model
from dbanu.utils.param import get_parsed_count_params, get_parsed_select_params
from dbanu.utils.string import to_var_name

Filter = TypeVar("Filter", bound=BaseModel)


def serve_select(
    app: FastAPI,
    query_engine: SelectEngine,
    select_query: str | Callable[[Filter], str],
    select_param: Callable[[Filter, int, int], list[Any]] | list[str] | None = None,
    count_query: str | Callable[[Filter], str] | None = None,
    count_param: Callable[[Filter], list[Any]] | list[str] | None = None,
    param: Callable[[Filter], list[Any]] | list[str] | None = None,
    path: str = "/get",
    methods: list[str] | None = None,
    filter_model: Type[BaseModel] | None = None,
    data_model: Type[BaseModel] | None = None,
    response_model: Type[BaseModel] | None = None,
    dependencies: list[Any] | None = None,
    middlewares: list[Middleware] | None = None,
    name: str | None = None,
    summary: str | None = None,
    description: str | None = None,
    default_limit: int | None = None,
):
    """
    Adding fastapi route to app with proper annotation:
    - taking filter_model as parameter
    - return pydantic model with two property:
        - data (data_model)
        - count (int)
    - supports FastAPI dependencies
    - supports middleware system
    """
    methods = ["GET"] if methods is None else [m.upper() for m in methods]
    var_name = to_var_name(name, path)
    if filter_model is None:
        filter_model = create_model(
            "FilterModel" if var_name is None else f"{var_name.capitalize()}Filter",
            limit=(int, default_limit if default_limit is not None else 100),
            offset=(int, 0),
        )
    if response_model is None:
        response_model = create_select_response_model(
            "ResponseModel" if var_name is None else f"{var_name.capitalize()}Response",
            data_model,
        )
    wrapped_dependencies = create_wrapped_fastapi_dependencies(dependencies)
    # Validate that all middlewares are async functions
    validate_middlewares(middlewares)

    # Create the actual handler function
    async def handle_request(
        request: Request,
        filter_data: filter_model,  # type: ignore
    ):
        """Select handler"""
        # Get params from filters
        limit = getattr(filter_data, "limit", 100)
        offset = getattr(filter_data, "offset", 0)
        # Extract dependency results from request state
        dependency_results = {}
        if request and hasattr(request.state, "dependency_results"):
            dependency_results = request.state.dependency_results
        # Build initial select parameters
        select_query_str = (
            select_query(filter_data) if callable(select_query) else select_query
        )
        parsed_select_params = get_parsed_select_params(
            filter_data, limit, offset, select_param, param
        )
        # Build initial count parameters
        count_query_str = (
            count_query(filter_data) if callable(count_query) else count_query
        )
        parsed_count_params = get_parsed_count_params(filter_data, count_param, param)
        # Create initial QueryContext
        initial_context = QueryContext(
            select_query=select_query_str,
            select_params=parsed_select_params,
            count_query=count_query_str,
            count_params=parsed_count_params,
            filters=filter_data,
            limit=limit,
            offset=offset,
            dependency_results=dependency_results,
        )
        query_processor = _create_query_processor(query_engine, response_model)
        handler = create_middleware_chain(middlewares, query_processor)
        return await handler(initial_context)

    # Register routes based on methods
    if "GET" in methods:

        @app.get(
            path,
            response_model=response_model,
            dependencies=wrapped_dependencies,
            name=None if name is None else f"{name}Get",
            summary=summary,
            description=description,
        )
        async def get_handler(
            request: Request,
            filters: filter_model = Depends(),  # type: ignore
        ):
            """Select route (generated by dbAnu)"""
            try:
                return await handle_request(request, filters)
            except Exception as e:
                traceback.print_exc()
                raise HTTPException(500, f"{e}")

    # Register other methods (POST, PUT, etc.)
    other_methods = [m for m in methods if m != "GET"]
    if other_methods:
        # Create a single route for all non-GET methods
        @app.api_route(
            path,
            methods=other_methods,
            response_model=response_model,
            dependencies=wrapped_dependencies,
            name=name,
            summary=summary,
            description=description,
        )
        async def non_get_handler(
            request: Request,
            filters: filter_model = Body(),  # type: ignore
        ):
            """Select route (generated by dbAnu)"""
            try:
                return await handle_request(request, filters)
            except Exception as e:
                traceback.print_exc()
                raise HTTPException(500, f"{e}")


def _create_query_processor(
    query_engine: SelectEngine, response_model: type[BaseModel]
):
    """Create a query processor for the middleware chain"""

    async def process_query(context: QueryContext):
        select_params = context.select_params or []
        # Check if select method is a coroutine
        select_method = query_engine.select
        if inspect.iscoroutinefunction(select_method):
            data = await select_method(context.select_query, *select_params)
        else:
            data = select_method(context.select_query, *select_params)
        if context.count_query:
            count_params = context.count_params or []
            # Check if select_count method is a coroutine
            select_count_method = query_engine.select_count
            if inspect.iscoroutinefunction(select_count_method):
                total = await select_count_method(context.count_query, *count_params)
            else:
                total = select_count_method(context.count_query, *count_params)
            return response_model(data=data, count=total)
        return response_model(data=data, count=len(data))

    return process_query



"""
DBAnu - FastAPI SQL Query Engine

A lightweight Python library that simplifies creating FastAPI endpoints for SQL queries.
"""

from dbanu.api import SelectSource, serve_select, serve_union
from dbanu.core import Middleware, QueryContext, SelectEngine
from dbanu.engines import MySQLQueryEngine, PostgreSQLQueryEngine, SQLiteQueryEngine

__all__ = [
    "MySQLQueryEngine",
    "PostgreSQLQueryEngine",
    "SQLiteQueryEngine",
    "QueryContext",
    "SelectEngine",
    "Middleware",
    "SelectSource",
    "serve_select",
    "serve_union",
]



"""
Middleware system for DBAnu
"""

import inspect
from typing import Any, Callable

from dbanu.core.engine import QueryContext

# Middleware type that receives QueryContext and next middleware callable
# and returns the ResponseModel (data_model)
Middleware = Callable[[QueryContext, Callable[[QueryContext], Any]], Any]


def validate_middlewares(middlewares: list[Middleware] | None) -> None:
    """Validate that all middleware functions are async"""
    if middlewares is None:
        return

    for i, middleware in enumerate(middlewares):
        if not inspect.iscoroutinefunction(middleware):
            raise TypeError(
                f"Middleware at index {i} must be an async function. "
                f"Got: {type(middleware).__name__}. "
                f"Please define your middleware with 'async def'."
            )


def create_middleware_chain(
    middlewares: list[Middleware] | None, final_handler: Callable[[QueryContext], Any]
):
    """Create a middleware chain from a list of middlewares"""
    handler = final_handler
    if middlewares is None:
        middlewares = []
    for middleware in reversed(middlewares):
        handler = _make_middleware_wrapper(middleware, handler)
    return handler


def _make_middleware_wrapper(
    current_middleware: Middleware, next_handler: Callable[[QueryContext], Any]
):
    """Wrap a middleware with the next handler"""

    async def wrapper(context: QueryContext):
        # Create an async-compatible next_handler
        async def async_next_handler(ctx):
            result = next_handler(ctx)
            if inspect.iscoroutine(result):
                return await result
            return result

        # All middleware should be async now
        return await current_middleware(context, async_next_handler)

    return wrapper


def get_combined_middlewares(
    global_middlewares: list[Middleware] | None,
    local_middlewares: list[Middleware] | None,
) -> list[Middleware] | None:
    """Combine global and local middlewares"""
    normalized_global_middlewares = global_middlewares or []
    normalized_local_middlewares = local_middlewares or []
    return list(normalized_global_middlewares) + list(normalized_local_middlewares)



"""
Database engines for DBAnu
"""

from dbanu.engines.mysql import MySQLQueryEngine
from dbanu.engines.postgresql import PostgreSQLQueryEngine
from dbanu.engines.sqlite import SQLiteQueryEngine

__all__ = ["SQLiteQueryEngine", "PostgreSQLQueryEngine", "MySQLQueryEngine"]
